#include <PalmOS.h>
#include "TDataSet.hpp"
#include "CommonFunc.hpp"
#define DB_TYPE 'DATA'

// Class TDataSet 
TDataSet::TDataSet()
{
	DB_NAME = NULL;
	DB_CREATOR = NULL;
	DB_CARDNO = 0;
	dbPtr = NULL;
	Position = 0;   //  
	FieldCount = 0; 
	LinkDS = NULL;
	MyDBPtr = true;
	FilterMembs = NULL;
	CountFltMemb = 0;
	FilteredRecs = NULL;
	Filtered = false;
	OnFuncValidateRecType = NULL;
	//OnGetFloatSizeField = NULL;
	HasFloatSizeFld = false;
	State = dsBrowse;
}

TDataSet::TDataSet(const TDataSet &right)
{
}


TDataSet::~TDataSet()
{
	TFilterMemb *next, *prev;
	int i;
	Close();
	for (i=0; i < FieldCount; ++i){
	  TField *field;
	  field = fields[i];
	  delete field;	
	}
	if (FilteredRecs) MemPtrFree(FilteredRecs); 
	if (FilterMembs){
	  prev = FilterMembs;
	  while (prev) { 
	    next = prev->next;
	    delete prev;
		prev = next;  
	  }
	}
}


TDataSet & TDataSet::operator=(const TDataSet &right)
{
}


int TDataSet::operator==(const TDataSet &right) const
{
}

int TDataSet::operator!=(const TDataSet &right) const
{
}

//     
UInt8 TDataSet::GetMemFldCount()
{  UInt8 i, count;
   
  count = 0;
  for(i = 0; i < FieldCount; ++i){
    if (fields[i]->isMemFld) ++count;
  }
  return count;
}


//## Other Operations (implementation)
Err TDataSet::Open (const DmOpenRef adbPtr=NULL)
{
	Char mess[50];
	UInt8 MemFldCount; //    
	
	
	
	if (dbPtr != NULL) return 0;
	
	if (adbPtr == NULL){
    	//Active = false;
		if ((DB_NAME == NULL)|( StrLen(DB_NAME)==0)) {
			SysFatalAlert ("   ");
			return 1; 
		}
		if ((DB_CREATOR == NULL)|( StrLen(DB_CREATOR)==0)) {
			SysFatalAlert ("  DB_CREATOR");
			return 1; 
		}

		// Open the database or create it if it doesn't exist:
		// First see if we can find it:
		LocalID dbID = DmFindDatabase(DB_CARDNO, DB_NAME);
		if(!dbID)
		{
			StrPrintF(mess, " '%s'  ", DB_NAME);
			SysFatalAlert (mess);
			return 1;
		}
		// Found it, now open it:
		dbPtr = DmOpenDatabase(DB_CARDNO, dbID, dmModeReadWrite);
	}
	else{
	  dbPtr = adbPtr;
      MyDBPtr = false;
	}   
	
	if(dbPtr == NULL) return 1;
	else if (GetMemFldCount() > 0){
        UInt8 i;
	    UInt32 aRecCount = RecCount();
	    //    
        for(i = 0; i < FieldCount; ++i){
          Char *p;		   
          if (fields[i]->isMemFld){
		    p = (Char *)MemPtrNew(fields[i]->get_DataLen() * aRecCount);
			fields[i]->MemData = p;
			ClearMem(fields[i]->MemData, fields[i]->get_DataLen() * aRecCount);
		  };
         }
	}
	return 0;  
}

UInt32 TDataSet::RecCount ()
{
  //## begin TDataSet::RecCount%3F7C29BD0179.body preserve=yes
  //## end TDataSet::RecCount%3F7C29BD0179.body
	if(dbPtr != NULL)
	{
	  if (Filtered) return(FltRecCount);
	  else 	return DmNumRecords(dbPtr);
	}else
	{
		Char mess[50];
		StrPrintF(mess, " '%s'  ", DB_NAME);
		SysFatalAlert (mess);
	}	 

}

void TDataSet::Close ()
{
  //## begin TDataSet::Close%3F7C2A2E0322.body preserve=yes
  //## end TDataSet::Close%3F7C2A2E0322.body
    if (dbPtr != NULL){
	    if (MyDBPtr)  DmCloseDatabase (dbPtr);
		//     MemFld
		if (GetMemFldCount() > 0){
           UInt8 i;
           for(i = 0; i < FieldCount; ++i)
             if ((fields[i]->isMemFld)&&(fields[i]->MemData)){
			   MemPtrFree(fields[i]->MemData);
			   fields[i]->MemData = NULL;
			 };
		}
		dbPtr = NULL;
	}	 
}

//  
Boolean TDataSet::Insert()
{
  UInt32 size, i;
   
  if (dbPtr == NULL) return false;
  
  size = 0;
  for(i = 0; i < FieldCount; ++i){
    if (!fields[i]->HasFloatSize && !fields[i]->isMemFld)
      size = size + fields[i]->get_DataLen(); 
  }
  
  Position = dmMaxRecordIndex;
  DmNewRecord(dbPtr, &Position, size);
  //Position = RecCount()-1;
  
  //        
  if (GetMemFldCount() > 0){
    UInt8 i;
	UInt32 aRecCount = RecCount();
	//    
    for(i = 0; i < FieldCount; ++i){
      Char *p;		   
      if (fields[i]->isMemFld){
	    UInt16 prevSize;
	    prevSize = MemPtrSize(fields[i]->MemData); 
	    MemPtrResize(fields[i]->MemData, fields[i]->get_DataLen() * aRecCount);
		ClearMem(fields[i]->MemData + prevSize, fields[i]->get_DataLen() * aRecCount - prevSize);
	  }
	}
  }
  return true;
}


void TDataSet::Drop ()
{
  Close();
  LocalID dbID = DmFindDatabase(DB_CARDNO, DB_NAME);
  DmDeleteDatabase (DB_CARDNO, dbID); 
}

void TDataSet::Clear ()
{
  
}

Boolean TDataSet::DBExists()
{
  LocalID dbID = DmFindDatabase(DB_CARDNO, DB_NAME);
  if(!dbID) return false;
  return true;
}

Err TDataSet::CreateDB ()
{
    LocalID dbID;
	Err err;
	
	Close();
	err = DmCreateDatabase(DB_CARDNO, DB_NAME, (UInt32)DB_CREATOR, DB_TYPE, false);
	if(err) return err;
	// Now we ought to find it:
	
	dbID = DmFindDatabase(DB_CARDNO, DB_NAME);
	if(!dbID)
		return 1;
	return Open ();	
  
}

UInt16 TDataSet::GetFilterPos(UInt16 pos)
{
  if (Filtered) {
    if (pos < FltRecCount) return FilteredRecs[pos];
	  else {
	    SysFatalAlert ("     ");  
	    return 0;
	  }		
  }else return pos;
}

MemHandle TDataSet::GetRecHandle(UInt16 Index)
{   Char mess[30];
	if(dbPtr != NULL)
	{
	    Index = GetFilterPos(Index);
	    return DmQueryRecord(dbPtr, Index);
	}
	else
	{
  	   StrPrintF(mess, " '%s'  ", DB_NAME);
	   SysFatalAlert (mess);
	}	 

}

TField* TDataSet::AddField (Char fieldname[30], TFldDataType fieldtype, 	//  :
  UInt32 datalen)
{   Int8 i;
    
	if (get_Active()){
	  SysFatalAlert ("     DataSet");
	  return NULL;
	}
	
	if (MaxFields < FieldCount){
		SysFatalAlert ("    ");
		return NULL;  
	}
	  fields[FieldCount] = new TField;
	  StrPrintF(fields[FieldCount]->FieldName, "%s", fieldname);
	  fields[FieldCount]->fieldtype = fieldtype;
	  fields[FieldCount]->DS = this;
	  fields[FieldCount]->set_DataLen(datalen);
	  fields[FieldCount]->MemData = NULL;
	  fields[FieldCount]->isMemFld = false;
	  fields[FieldCount]->Editable = false;
	  fields[FieldCount]->index = FieldCount; 
	  
	  if (FieldCount == 0) fields[FieldCount]->set_offset(0);
	  else{
	    i = FieldCount - 1;
	    //        
	    while (i>=0){
	      if (!fields[i]->isMemFld) break;
		  --i;
		}  
	    if (i < 0) fields[FieldCount]->set_offset(0);
		else{
  	      fields[FieldCount]->set_offset( 
	        fields[i]->get_offset()  + fields[i]->get_DataLen());
		}  
	  }
      FieldCount++;
	  return fields[FieldCount - 1];
}  

//          ,   Open(),   Close()	  
TField* TDataSet::AddMemField (Char fieldname[30], TFldDataType fieldtype, 	//  :
  UInt32 datalen)
{  TField *newFld;

  newFld = AddField (fieldname,fieldtype, datalen);
  if (newFld) {
    newFld->isMemFld = true;
	newFld->set_offset(0);
	newFld->Editable = true;
  }
  else 
    return NULL; 
  return fields[FieldCount - 1];
}  

//        fldSize  	  
TField* TDataSet::AddFloatSizeField (Char fieldname[30], 
    	TFldDataType fieldtype, 	//  :
		TField *fldSize)
{  TField *newFld;

  HasFloatSizeFld = false; //     
  newFld = AddField (fieldname, fieldtype, 0);
  if (newFld) {
    newFld->HasFloatSize = true;
	newFld->FldHaveSize = fldSize;
	HasFloatSizeFld = true;
  }
  else 
    return NULL; 
  return fields[FieldCount - 1];
}  


Boolean TDataSet::LinkPositioning() //    TDataSet
{ 
  if (LinkDS != NULL) 
    if (!LocateByField(LinkField, LinkDS->GetFieldData(LinkDSField)))
	  return false;
  return true;	  
	  //SysFatalAlert ("     ");  
}


/*Char* TDataSet::GetFieldDataByName(Char fieldname[30], const Boolean positing=true)
{
  UInt32 i, readpos;
  Char *p;
  Char mess[50];
  TField field;
  MemHandle h;
  Char *n;
  
  field = GetFieldByName(fieldname);
  if (positing) //   DS
    if (!LinkPositioning()){
	  return NULL;
    }
  h = GetRecHandle(Position);
  n = (Char *) MemHandleLock(h);
  //p = (Char *)MemPtrNew(fields[i]->get_DataLen());
  MemHandleUnlock(h);
  //StrNCopy (p, n + fields[i]->get_offset(), fields[i]->get_DataLen());
  return n + field.offset; 
}*/

Char* TDataSet::GetFieldData(TField *field, const Boolean positing=true)
{
  MemHandle h;
  Char *n;
  
  if (positing) //   DS
    if (!LinkPositioning()){
      SysFatalAlert ("  ");
	  return NULL;
    }
  if (field->isMemFld){
    if (field->MemData){
	  UInt16 pos = GetFilterPos(Position); 
	  if (MemPtrSize(field->MemData) > (pos * field->get_DataLen())){
        n = &field->MemData[pos * field->get_DataLen()];
	    return n;
	  }
	  else{
	    SysFatalAlert ("        ");
		return NULL;
	  }	
	}  
	else{
	  SysFatalAlert ("    MemField  ");
	  return NULL;
	}
  }
  else{	
    h = GetRecHandle(Position);
    n = (Char *) MemHandleLock(h);
    MemHandleUnlock(h);
    return n + field->get_offset();
  }	 
}
//  
Err TDataSet::SetFieldData(TField *field, Char *p, const Boolean positing=true)
{
  MemHandle h;
  Char *n;
  
  if (positing) //   DS
    if (!LinkPositioning()){
      SysFatalAlert ("  ");
	  return dmErrCantFind;
    }
  if (field->isMemFld){
    if (field->MemData){
      UInt16 pos = GetFilterPos(Position); 
	  if (MemPtrSize(field->MemData) > (pos * field->get_DataLen())){
	    n = &field->MemData[pos * field->get_DataLen()];
	    StrNCopy (n, p, field->get_DataLen());
	    return errNone;
	  }
	  else{
	    SysFatalAlert ("        ");
		return dmErrIndexOutOfRange;
	  }	
	}  
	else{
	  SysFatalAlert ("    MemField  ");
	  return dmErrResourceNotFound;
	}
  }
  else
  if (field->HasFloatSize){
    int i;
    UInt32 size, rsize;
	Char *storeRec;
	
	//    
	//if (field->get_DataLen() == 0) return errNone;
    size = 0;
    for(i = 0; i < FieldCount; ++i){
      if (!fields[i]->isMemFld)
        size = size + fields[i]->get_DataLen(); 
    }
	
	//           
	h = GetRecHandle(Position);
	rsize = MemHandleSize(h);
	if (rsize != size){
	  //   
	  DmResizeRecord (dbPtr, Position, size);
	} 
	if (field->get_DataLen() != 0){
      n = (Char *) MemHandleLock(h);
	  DmWrite(n, field->get_offset(), p, field->get_DataLen());
      MemHandleUnlock(h);
	}  
    return errNone;
  }
  else{	
    UInt16 offset, len; 
    h = GetRecHandle(Position);
    n = (Char *) MemHandleLock(h);
	offset = field->get_offset();
	//n = n + offset; 
	len = field->get_DataLen();
	
	switch (DmWriteCheck(n, offset, len))
	{
		case errNone:
		    DmWrite(n, offset, p, len);
			break;
		case dmErrNotValidRecord:
	        SysFatalAlert ("ErrNotValidRecord");
			break;
		case dmErrWriteOutOfBounds:
	        SysFatalAlert ("ErrWriteOutOfBounds");
			break;
		default:
			break;
	}
	
    //StrNCopy (n, p, len);
    MemHandleUnlock(h);
    return errNone;
  }	 
}



MemHandle TDataSet::GetFieldDataHandle(TField *field, Char **p)
{
  //UInt32 i;
  //TField field;
  Char *n;
  MemHandle h;
  
  if (field->isMemFld){
	SysFatalAlert ("GetFieldDataHandle    MemField  ");
	return NULL;
  }	
    
  if (!LinkPositioning())
    {  
       return NULL;
	}   
  
  //field = GetFieldByName(fieldname);
  
  h = GetRecHandle(Position);
  n = (Char *) MemHandleLock(h);
  *p = n + field->get_offset(); 
  //p = (Char *)MemPtrNew(fields[i]->get_DataLen());
  MemHandleUnlock(h);
  //StrNCopy (p, n + fields[i]->get_offset(), fields[i]->get_DataLen());
  return h; 
}


TField *TDataSet::GetFieldByName(Char fieldname[30])
{
  UInt32 i;
  Char mess[50];
  
  for(i = 0; i < FieldCount; ++i){
    if (StrCompare(fieldname,fields[i]->FieldName) == 0) return fields[i];
  }
  StrPrintF(mess, " '%s'  ", fieldname);
  SysFatalAlert (mess);
}

Boolean TDataSet::LocateByField(TField *field, Char *data)
{ UInt16 SavePos, i, RecordCount;
  Char *p;
  //TField field;
  

  //field = GetFieldByName(fieldname);
  //    
  p = GetFieldData(field, false);
  if (MemCmp(p, data, field->get_DataLen()) == 0) return true;

  SavePos = Position;
  RecordCount = RecCount();
  for(i = 0; i < RecordCount; ++i){
    Position = i;
	p = GetFieldData(field, false);
	/*if (field.DataLen == 4){
	  UInt32 int1 = *(UInt32 *)p; 
	  UInt32 int2 = *(UInt32 *)data;
	  if (int1==int2) return true;
	}
	else*/
    if (MemCmp(p, data, field->get_DataLen()) == 0) return true;
    //if (StrNCompare(p, data, field.DataLen) == 0) return true;
  }
  Position = SavePos;
  return false;
}

//       "p" 
Char* TDataSet::FieldAsPChar(TField *field, Char *p)
{
  UInt32 len;
  Boolean postrue;
  
  if (!LinkPositioning()){
    StrCopy (p, "???");
	p[4] = 0;
	return p;
  };
  switch (field->fieldtype){
	case fdtChar : { //     ..        
	   Char *n;
	   n = GetFieldData(field, false);
	   len = StrLen(n);
       StrNCopy (p, n, len);
	   p[len] = 0;
       return p;
	   break;
    }
    case fdtInt : {
      switch (field->get_DataLen()){
	    case 1 : { 
  	      UInt8 *lInt;
	      lInt = (UInt8 *)GetFieldData(field, false);
	      StrIToA(p, *lInt);
		  break; 
	    }
	    case 2 : { 
  	      UInt16 *lInt;
	      lInt = (UInt16 *)GetFieldData(field, false);
	      StrIToA(p, *lInt);
		  break; 
	    }
	    case 4 : { 
  	      UInt32 *lInt;
	      lInt = (UInt32 *)GetFieldData(field, false);
	      StrIToA(p, *lInt);
		  break; 
	    }
	  }	
	  
	  return p;
  	  break;
	}
    case fdtCurrency : {
      FlpCompFloat *lFloat;
	  lFloat = (FlpCompFloat *)GetFieldData(field, false);
	  FloatToCurrency(*lFloat, p, 2);  
	  return p;
  	  break;
    }
    case fdtDate : {
	  MyDatePtr dt;
	  dt = (MyDatePtr)(GetFieldData(field, false));
	  DateToStr(dt->day, dt->month, dt->year, p);
	  p[6] = p[8]; 
	  p[7] = p[9];
	  p[8] = 0; 
	  return p;
  	  break;
    }
    default: 
	  SysFatalAlert ("   ");  
  }
  
}

Boolean TDataSet::SetLinkDS(TField *linkDSField, TField *linkField)
{
  if (linkDSField->DS == this) 
	SysFatalAlert ("  DataSet  ");
  else
    {	
      LinkField = linkField;
      LinkDSField = linkDSField;
      LinkDS = LinkDSField->DS;
	}  
}

Boolean TDataSet::ClearLink(void)
{
  LinkField = NULL;
  LinkDSField = NULL;
  LinkDS = NULL;
}


TFilterMemb* TDataSet::AddFilterMemb(TField *afield, Char *adata, TEqType aEqType)
{  
  TFilterMemb *last, *cur;
  
  cur = new TFilterMemb;
  if (CountFltMemb > 0){
    last = FilterMembs;
    while (last->next) 
	  last = last->next;
	last->next = cur; 
  }
  else FilterMembs = cur;
  cur->data = adata;
  cur->EqType = aEqType;
  cur->field = afield;
  cur->next = NULL;
  ++CountFltMemb;
  return cur;
   
}

void TDataSet::set_Filtered (Boolean value)
{  UInt32 RecordCount, 
     i, j;
  TFilterMemb *cur, *next;	
  Boolean Valid;
  Char *p;
  
  Boolean isAdd;
  
  if ((!value)||(Filtered)){
    if (FilteredRecs) MemPtrFree(FilteredRecs);
	FilteredRecs = NULL;
    Filtered = value;
	FltRecCount = -1;
	if (!value) return;
  }
  
  RecordCount = RecCount();
  FilteredRecs = (UInt16 *)MemPtrNew(RecordCount * sizeof(UInt16));
   
  FltRecCount = 0; 
   //        
  for(i = 0; i < RecordCount; ++i){
    Position = i;
    isAdd = false;
    cur = FilterMembs;
	//    
	if (OnFuncValidateRecType) {
	  isAdd = OnFuncValidateRecType(this); 
	}
	//    
	else
	while (cur && !isAdd) {
	  p = GetFieldData(cur->field, false);
	  switch (cur->EqType){
        case etEqual : //
           if (MemCmp(p, cur->data, cur->field->get_DataLen()) == 0) {
	 	     isAdd = true;
		 	 continue;
		   }	
        /*case etMore : // 
		   //  
		   if (cur->field->fieldtype == fdtDate){
			 DateTimePtr dt;
	         dt = (DateTimePtr)(p - 6);
	 	     isAdd = true;
		   }
		   else
           if (MemCmp(p, cur->data, cur->field->get_DataLen()) == 0) {
		 	 continue;
		   }	*/
	  }	
      cur = cur->next;
	} 
	if (isAdd){ //    
	   FilteredRecs[FltRecCount] = i;
  	   ++FltRecCount; 
	} 
  }
  MemPtrResize(FilteredRecs, FltRecCount*sizeof(UInt16));
  Filtered = value;
} 

void TDataSet::set_Position (UInt16 value)
{
  if (RecCount() > value)
    Position = value;
  else
	SysFatalAlert ("     ");  
}
